/**
 * Copyright (C) 2009-2010 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif

#include <glib/gstdio.h>
#include <gio/gio.h>

#include <gdk/gdkkeysyms.h>

#include <libsoup/soup.h>

#include "utils.h"

void
seconds_to_string (gint seconds, gchar **str)
{
    gint minutes, hours;
    gchar *tmp = NULL;

    if (G_UNLIKELY (seconds <=0))
        tmp = g_strdup ("00:00");
    else
    {
        minutes = seconds / 60;
        seconds = seconds % 60;

        if (minutes >= 60) {
            hours = minutes / 60;
            minutes = minutes % 60;
            tmp = g_strdup_printf ("%d:%02d:%02d", hours, minutes, seconds);
        }
        else
            tmp = g_strdup_printf ("%02d:%02d", minutes, seconds);
    }

    g_free (*str);
    *str = tmp;
}

void
expand_uri (gchar **uri)
{
    if (*uri == NULL)
        return;

    gchar *tmp = NULL;

    if (! g_regex_match_simple ("^[a-z]*://", *uri, G_REGEX_CASELESS, 0))
        tmp = g_strconcat ("http://", *uri, NULL);

    if (tmp)
    {
        g_free (*uri);
        *uri = tmp;
    }

}

void
empty_dir (const gchar *dirname)
{
    GDir *dir;
    const gchar *name;

    dir = g_dir_open (dirname, 0, NULL);
    if (!dir)
        return;

    while ((name = g_dir_read_name (dir)) != NULL)
    {
        gchar *path = g_build_filename (dirname, name, NULL);
        if (g_file_test (path, G_FILE_TEST_IS_DIR))
        {
            empty_dir (path);
            g_rmdir (path);
        }
        else
        {
            g_remove (path);
        }
        g_free (path);
        path = NULL;
    }
    g_dir_close (dir);
}

void
kill_external (pid_t pid)
{
    kill (pid, SIGTERM);
    usleep (50000);
    if (kill (pid, 0) != 0)
        kill (pid, SIGKILL);
}

/*
 * From exo (Xfce)
 * Copyright (c) 2004-2007 os-cillation e.K.
 */
gchar *
exo_str_replace (const gchar *str,
                 const gchar *pattern,
                 const gchar *replacement)
{
    const gchar *s, *p;
    GString         *result;

    g_return_val_if_fail (str != NULL, NULL);
    g_return_val_if_fail (pattern != NULL, NULL);
    g_return_val_if_fail (replacement != NULL, NULL);

    /* empty patterns are kinda useless, so we just return a copy of str */
    if (G_UNLIKELY (*pattern == '\0'))
        return g_strdup (str);

    /* allocate the result string */
    result = g_string_new (NULL);

    /* process the input string */
    while (*str != '\0')
    {
        if (G_UNLIKELY (*str == *pattern))
        {
            /* compare the pattern to the current string */
            for (p = pattern + 1, s = str + 1; *p == *s; ++s, ++p)
                if (*p == '\0' || *s == '\0')
                    break;

            /* check if the pattern matches */
            if (G_LIKELY (*p == '\0'))
            {
                g_string_append (result, replacement);
                str = s;
                continue;
            }
        }

        g_string_append_c (result, *str++);
    }

    return g_string_free (result, FALSE);
}

gchar *
wildcard_to_regex (const gchar *input)
{
    gchar *tmp = NULL;
    gchar *ret = NULL;

    if (g_utf8_strlen (input, -1) == 0)
        return NULL;

    tmp = exo_str_replace (input, ".", "\\.");
    ret = exo_str_replace (tmp, "*", ".*");

    g_free (tmp);

    return ret;
}

gchar *
get_remote_data (const gchar *url)
{
    gchar *data = NULL;
    SoupSession *session;
    SoupMessage *msg;

    session = soup_session_sync_new ();
    msg = soup_message_new ("GET", url);

    if (soup_session_send_message (session, msg) == SOUP_STATUS_OK)
        data = g_strdup (msg->response_body->data);

    g_object_unref (msg);
    g_object_unref (session);

    return data;
}

static const char letters[] = "abcdefghijklmnopqrstuvwxyz"
                              "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                              "0123456789";
static gchar *
gen_random_str (guint len)
{
    guint i;
    gchar *ret;
    ret = (gchar *) g_malloc (len+1);

    for (i=0;i<len;i++)
        ret[i] = letters[rand() % (sizeof letters - 1)];

    return ret;
}

gchar *
crypt_passwd (const gchar *passwd, const gchar *given_salt)
{
    gchar *rnd = NULL;
    gchar *salt = NULL;
    gchar *crypted_pass = NULL;

    if (! given_salt)
    {
        rnd = gen_random_str (6);
        salt = g_strdup_printf ("$1$%s", rnd);
        g_free (rnd);
    }
    else
    {
        salt = g_strndup (given_salt, 10);
    }

    crypted_pass = g_strdup (crypt (passwd, salt));

    g_free (salt);

    return crypted_pass;
}

gboolean
validate_passwd (const gchar *passwd, const gchar *crypted)
{
    gchar *computed_passwd = NULL;
    gboolean ret;

    if (!passwd || !crypted)
        return FALSE;

    computed_passwd = crypt_passwd (passwd, crypted);
    ret = ((g_strcmp0 (computed_passwd, crypted)) == 0) ? TRUE : FALSE;

    g_free (computed_passwd);

    return ret;
}

GPid unclutter_pid = 0;
gchar *unclutter_args[] =
{
    "unclutter",
    "-idle", "1",
    NULL
};

void
start_unclutter ()
{
    static gboolean started = FALSE;

    if (started)
        return;

    g_spawn_async (NULL, unclutter_args, NULL, G_SPAWN_SEARCH_PATH,
                   NULL, NULL, &unclutter_pid, NULL);
}

void
stop_unclutter ()
{
    if (!unclutter_pid)
        return;

    kill (unclutter_pid, SIGTERM);
    g_spawn_close_pid (unclutter_pid);
    unclutter_pid = 0;
}

gboolean
create_dir_for_file (const gchar *filename)
{
    int ret;
    gchar *dir = NULL;

    dir = g_path_get_dirname (filename);
    ret = g_mkdir_with_parents (dir, 0755);
    g_free (dir);

    return (ret == 0);
}

gboolean
is_local_image_uri (const gchar *uri)
{
    gboolean certain;
    gboolean is_image;
    gchar *type;

    g_return_val_if_fail (uri != NULL, FALSE);

    if (!g_str_has_prefix (uri, "file://"))
        return FALSE;

    type = g_content_type_guess (uri+7, NULL, 0, &certain);

    is_image = g_str_has_prefix (type, "image/");

    g_free (type);

    return is_image;
}

gboolean
is_video_uri (const gchar *uri)
{
    gboolean certain;
    gboolean is_video;
    gchar *type;

    g_return_val_if_fail (uri != NULL, FALSE);

    type = g_content_type_guess (uri+7, NULL, 0, &certain);

    is_video = g_str_has_prefix (type, "video/");

    g_free (type);

    return is_video;
}

gint
exec_hook (const gchar *hook_dir, const gchar *hook)
{
    gchar *script;
    gchar *out = NULL, *err = NULL;
    gint status = -1;

    if (!hook_dir || !hook)
        return status;

    script = g_strdup_printf ("%s/%s", hook_dir, hook);
    if (g_file_test (script, G_FILE_TEST_IS_EXECUTABLE))
        g_spawn_command_line_sync (script, &out, &err, &status, NULL);

    g_free (out);
    g_free (err);
    g_free (script);

    return status;
}

