/**
 * Copyright (C) 2011 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <stdio.h>
#include <string.h>

#include <curl/curl.h>

#include "main.h"
#include "wk-download.h"

static CURL *curl = NULL;

static gboolean
init_curl ()
{
    if ((curl = curl_easy_init ()) == NULL)
        return FALSE;

    curl_easy_setopt (curl, CURLOPT_FAILONERROR, 1);
    curl_easy_setopt (curl, CURLOPT_USERAGENT, "linutop-wkdaemon/" PACKAGE_VERSION);
    curl_easy_setopt (curl, CURLOPT_NOPROGRESS , 1);

    return TRUE;
}

static gchar *
get_etag ()
{
    time_t filetime = 0;
    double filesize = 0;

    curl_easy_getinfo (curl, CURLINFO_FILETIME, &filetime);
    curl_easy_getinfo (curl, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &filesize);

    if (!filetime || !filesize)
        return NULL;

    return g_strdup_printf ("%ld:%.0f", filetime, filesize);
}

gboolean
wk_download_is_needed (const gchar *source, const gchar *target)
{
    gchar *etag_file = NULL;
    gchar *etag_local = NULL;
    gchar *etag_distant = NULL;
    gboolean ret;

    if (!curl && !init_curl ())
        return FALSE;

    etag_file = g_strdup_printf ("%s.etag", target);
    if (!g_file_test (etag_file, G_FILE_TEST_EXISTS))
    {
        ret = TRUE;
        goto end;
    }

    if (!g_file_get_contents (etag_file, &etag_local, NULL, NULL))
    {
        ret = TRUE;
        goto end;
    }

    curl_easy_setopt (curl, CURLOPT_URL, source);
    curl_easy_setopt (curl, CURLOPT_NOBODY, 1L);
    curl_easy_setopt (curl, CURLOPT_FILETIME, 1L);
    curl_easy_setopt (curl, CURLOPT_HEADER, 0L);

    if (curl_easy_perform (curl) != CURLE_OK)
    {
        ret = TRUE;
        goto end;
    }

    etag_distant = get_etag ();
    if (! etag_distant)
    {
        ret = TRUE;
        goto end;
    }

    DEBUG ("etag_local: %s", etag_local);
    DEBUG ("etag_distant: %s", etag_distant);

    if (strcmp (etag_local, etag_distant) == 0)
    {
        ret = FALSE;
    }
    else
    {
        ret = TRUE;
        g_file_set_contents (etag_file, etag_distant, -1, NULL);
    }
    g_free (etag_distant);

end:
    g_free (etag_file);
    g_free (etag_local);

    return ret;
}




gboolean
wk_download_perform (const gchar *source, const gchar *target)
{
    FILE *fp;
    gchar *etag;
    gchar *etag_file;

    if (!curl && !init_curl ())
        return FALSE;

    fp = fopen (target, "w");
    if (! fp)
        return FALSE;

    curl_easy_setopt (curl, CURLOPT_URL, source);
    curl_easy_setopt (curl, CURLOPT_NOBODY, 0L);
    curl_easy_setopt (curl, CURLOPT_FILETIME, 0L);
    curl_easy_setopt (curl, CURLOPT_WRITEDATA, fp);
    curl_easy_setopt (curl, CURLOPT_FILETIME, 1L);

    if (curl_easy_perform (curl) != CURLE_OK)
        return FALSE;

    etag = get_etag ();
    etag_file = g_strdup_printf ("%s.etag", target);
    g_file_set_contents (etag_file, etag, -1, NULL);
    g_free (etag_file);
    g_free (etag);

    fclose (fp);

    return TRUE;
}

