/**
 * Copyright (C) 2010 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include "wk-pl-parser.h"
#include "wk-slideshow-item.h"

#define DEFAULT_ITEM_TIMEOUT 5

static int cur_priority;
static int cur_id;
static gboolean in_loop = FALSE;

typedef struct {
    int start_year;
    int start_month;
    int start_day;
    int start_hour;
    int start_min;

    int end_year;
    int end_month;
    int end_day;
    int end_hour;
    int end_min;

    int weekdays;
} WkPlLoopInfo;


static void
append_dir_to_list (const gchar *dirname, gint t, GSList **l)
{
    GDir *dir;
    gchar *filename = NULL, *uri = NULL;
    const gchar *localname = NULL;
    WkSlideshowItem *item;
    WkSlideshowItemType type;

    if ((dir = g_dir_open (dirname, 0, NULL)) == NULL)
        return;

    while ((localname = g_dir_read_name (dir)) != NULL)
    {
        uri = g_strdup_printf ("file://%s/%s", dirname, localname);
        filename = uri+7;

        if (g_file_test (filename, G_FILE_TEST_IS_DIR))
        {
            append_dir_to_list (filename, t, l);
            g_free (uri);
        }
        else
        {
            type = wk_slideshow_item_guess_type (uri);
            item = wk_slideshow_item_new (uri, type, t);
            g_free (uri);
            *l = g_slist_append (*l, item);
        }
    }
    g_dir_close (dir);
}

GSList *
old_school_playlist_parser (const gchar *str)
{
    WkSlideshowItem *item;
    WkSlideshowItemType type;
    GSList *playlist = NULL;
    gchar **lines = NULL;
    gchar **p;
    gchar buf[512];
    int t;

    lines = g_strsplit (str, "\n", 0);
    for (p=lines; *p; p++)
    {
        if (sscanf (*p, "%s %d", buf, &t) == 2)
        {
            if (g_file_test (buf+7, G_FILE_TEST_IS_DIR))
            {
                append_dir_to_list (buf+7, t, &playlist);
                continue;
            }

            type = wk_slideshow_item_guess_type (buf);
            item = wk_slideshow_item_new (buf, type, t);
            playlist = g_slist_append (playlist, item);
        }
    }
    g_strfreev (lines);

    return playlist;
}

gboolean
wk_pl_loop_get_from_markup (const char *start,
                            const char *end,
                            const char *weekdays,
                            WkPlLoopInfo *info)
{
    char *p;

    info->start_year = -1;
    info->end_year = -1;
    info->start_hour = -1;
    info->end_hour = -1;
    info->weekdays = -1;

    if (weekdays)
    {
        info->weekdays = 0;
        for (p=(char *)weekdays;*p;p++)
        {
            if (*p >= '0' && *p <= '6')
                info->weekdays |= 1 << (*p - '0');
        }
    }

    /* if 'start' is NULL, we don't care about 'end' */
    if (!start)
        return (info->weekdays >= 1) ? TRUE : FALSE;

    if (strlen (start) > 5) /* full date */
    {
        if (5 != sscanf (start, "%d/%d/%d %d:%d", &info->start_year,
                                                  &info->start_month,
                                                  &info->start_day,
                                                  &info->start_hour,
                                                  &info->start_min))
        {
            return FALSE;
        }

        if (end)
        {
            if (5 != sscanf (end, "%d/%d/%d %d:%d", &info->end_year,
                                                    &info->end_month,
                                                    &info->end_day,
                                                    &info->end_hour,
                                                    &info->end_min))
            {
                return FALSE;
            }
        }

        return TRUE;
    }
    else
    {
        if (2 != sscanf (start, "%d:%d", &info->start_hour, &info->start_min))
            return FALSE;

        if (2 != sscanf (end, "%d:%d", &info->end_hour, &info->end_min))
            return FALSE;

        return TRUE;
    }
}

void
wk_pl_info_print (WkPlLoopInfo *info)
{
    int i;

    if (info->start_year > 0)
    {
        printf ("Start: %d/%d/%d %d:%d\n", info->start_year,
                                           info->start_month,
                                           info->start_day,
                                           info->start_hour,
                                           info->start_min);
        if (info->end_year > 0)
            printf ("End  : %d/%d/%d %d:%d\n", info->end_year,
                                               info->end_month,
                                               info->end_day,
                                               info->end_hour,
                                               info->end_min);
    }
    else
    {
        if (info->start_hour >= 0)
        {
            printf ("Start: %d:%d\n", info->start_hour, info->start_min);
            printf ("End  : %d:%d\n", info->end_hour, info->end_min);
        }
    }

    if (info->weekdays > 0)
    {
        printf ("Days : %d (", info->weekdays);
        for (i=0;i<7;i++)
            if (info->weekdays & 1<<i)
                printf ("%d ", i);
        printf ("\b)\n");
    }
}

int
wk_pl_loop_detail_level (WkPlLoopInfo *info)
{
    int ret = 0;
    ret += (info->weekdays > 0)    ? 1 : 0;
    ret += (info->start_hour >= 0) ? 1 : 0;
    ret += (info->start_year >= 0) ? 1 : 0;

    return ret;
}

void
wk_pl_loop_to_struct_tm (WkPlLoopInfo *info,
                         struct tm *tm_start,
                         struct tm *tm_end,
                         struct tm *now)
{
    if (info->start_hour >= 0)
    {
        tm_start->tm_sec = 0;
        tm_start->tm_min = info->start_min;
        tm_start->tm_hour = info->start_hour;

        tm_end->tm_sec = 0;
        tm_end->tm_min = info->end_min;
        tm_end->tm_hour = info->end_hour;
    }

    if (info->start_year > 0)
    {
        tm_start->tm_mday = info->start_day;
        tm_start->tm_mon = info->start_month - 1;
        tm_start->tm_year = info->start_year - 1900;

        if (info->end_year > 0)
        {
            tm_end->tm_mday = info->end_day;
            tm_end->tm_mon = info->end_month - 1;
            tm_end->tm_year = info->end_year - 1900;
        }
        /* We will not use tm_end if end_year is not defined */
    }
    else
    {
        tm_start->tm_mday = now->tm_mday;
        tm_start->tm_mon = now->tm_mon;
        tm_start->tm_year = now->tm_year;

        tm_end->tm_mday = now->tm_mday;
        tm_end->tm_mon = now->tm_mon;
        tm_end->tm_year = now->tm_year;
    }
}

gboolean
wk_pl_loop_fits_now (WkPlLoopInfo *info)
{
    time_t t_now;
    struct tm *now;

    struct tm tm_start;
    struct tm tm_end;

    t_now = time (NULL);
    now = localtime (&t_now);

    /* we build struct tm's from the info to easily do calculations using the
     * epoch */
    wk_pl_loop_to_struct_tm (info, &tm_start, &tm_end, now);

    /* If a date is defined */
    if (info->start_year > 0)
    {
        /* if weekdays is defined and today is not in the list, we bail */
        if ((info->weekdays >= 0 ) && (! (info->weekdays & (1 << now->tm_wday))))
            return FALSE;

        /* end is also defined => now is included between start and end? */
        if (info->end_year > 0)
        {
            if ((t_now >= mktime (&tm_start)) && (t_now <= mktime (&tm_end)))
                return TRUE;
        }
        else /* end undefined => now after start? */
        {
            if (t_now >= mktime (&tm_start))
                return TRUE;
        }
    }
    else
    {
        /* if weekdays is defined and today is not in the list, we bail */
        if ((info->weekdays >= 0 ) && (! (info->weekdays & (1 << now->tm_wday))))
            return FALSE;

        /* if the week day is OK, test the hours */
        if (info->start_hour >= 0)
        {
            if ((t_now >= mktime (&tm_start)) && (t_now <= mktime (&tm_end)))
                return TRUE;
        }
        else
        {
            return TRUE;
        }
    }

    return FALSE;
}

static void
start_element (GMarkupParseContext *context,
               const gchar         *element_name,
               const gchar        **attribute_names,
               const gchar        **attribute_values,
               gpointer             user_data,
               GError             **error)
{
    const gchar *start = NULL;;
    const gchar *end = NULL;
    const gchar *weekdays = NULL;
    const gchar *id = NULL;

    WkPlLoopInfo info;

    int i, prio;

    if (strcmp (element_name, "loop") != 0)
        return;

    for (i=0; attribute_names[i] != NULL; i++)
    {
        if (0 == strcmp (attribute_names[i], "start"))
            start = attribute_values[i];
        else if (0 == strcmp (attribute_names[i], "end"))
            end = attribute_values[i];
        else if (0 == strcmp (attribute_names[i], "weekdays"))
            weekdays = attribute_values[i];
        else if (0 == strcmp (attribute_names[i], "id"))
            id = attribute_values[i];
    }

    if (!id)
        return;

    if (! wk_pl_loop_get_from_markup (start, end, weekdays, &info))
        return;

    prio = wk_pl_loop_detail_level (&info);
    if (prio > cur_priority)
    {
        cur_priority = prio;
        cur_id = atoi (id);
    }
}

static const GMarkupParser
first_run_parser = {
    start_element,
    NULL,
    NULL,
    NULL,
    NULL
};

static void
start_element_list (GMarkupParseContext *context,
                    const gchar         *element_name,
                    const gchar        **attribute_names,
                    const gchar        **attribute_values,
                    gpointer             user_data,
                    GError             **error)
{
    int i, id = -1;
    const gchar *uri = NULL;
    WkSlideshowItemType type = WK_SLIDESHOW_ITEM_TYPE_UNKNOWN;
    int t = 0;
    GSList **l = (GSList **) user_data;
    GSList *tmp;

    if (! in_loop)
    {
        if (strcmp (element_name, "loop") != 0)
            return;

        for (i=0; attribute_names[i] != NULL; i++)
        {
            if (0 == strcmp (attribute_names[i], "id"))
            {
                id = atoi (attribute_values[i]);
                break;
            }
        }
        if (id != cur_id)
            return;

        in_loop = TRUE;
        return;
    }

    for (i=0; attribute_names[i] != NULL; i++)
    {
        if (0 == strcmp (attribute_names[i], "uri"))
        {
            uri = attribute_values[i];
        }
        else if (0 == strcmp (attribute_names[i], "time"))
        {
            t = atoi (attribute_values[i]);
        }
        else if (0 == strcmp (attribute_names[i], "type"))
        {
            if (strcmp (attribute_values[i], "image") == 0)
                type = WK_SLIDESHOW_ITEM_TYPE_IMAGE;
            else if (strcmp (attribute_values[i], "video") == 0)
                type = WK_SLIDESHOW_ITEM_TYPE_VIDEO;
            else
                type = WK_SLIDESHOW_ITEM_TYPE_WEB;
        }
    }

    if (!uri)
        return;

    if (type == WK_SLIDESHOW_ITEM_TYPE_UNKNOWN)
        type = wk_slideshow_item_guess_type (uri);

    if (!t)
        t= DEFAULT_ITEM_TIMEOUT;

    if (uri && t)
    {
        tmp = g_slist_append (*l, wk_slideshow_item_new (uri, type, t));
        *l = tmp;
    }
}

static void
end_element_list (GMarkupParseContext *context,
                  const gchar         *element_name,
                  gpointer             user_data,
                  GError             **error)
{
    if (in_loop && (0 == strcmp (element_name, "loop")))
    {
        in_loop = FALSE;
    }
}

static const GMarkupParser
second_run_parser = {
    start_element_list,
    end_element_list,
    NULL,
    NULL,
    NULL
};

void
wk_pl_free_playlist_item (gpointer data, gpointer userdata)
{
    wk_slideshow_item_free ((WkSlideshowItem *) data);
}

void
wk_pl_free_playlist (GSList *playlist)
{
    g_slist_foreach (playlist, wk_pl_free_playlist_item, NULL);
    g_slist_free (playlist);
}

GSList *
wk_pl_get_playlist (const gchar *xml)
{
    GMarkupParseContext *context;
    GSList *playlist = NULL;

    cur_priority = 0;
    cur_id = 0;

    /* first parsing defines which loop id to use */
    context = g_markup_parse_context_new (&first_run_parser, 0, NULL, NULL);
    if (!g_markup_parse_context_parse (context, xml, strlen (xml), NULL))
    {
        g_markup_parse_context_free (context);

        /* Use the "old" syntax (uri timeout) */
        playlist = old_school_playlist_parser (xml);
        goto end;
    }

    /* second parsing gets the actual playlist */
    context = g_markup_parse_context_new (&second_run_parser, 0, &playlist, NULL);
    g_markup_parse_context_parse (context, xml, strlen (xml), NULL);

end:
    return playlist;
}

