/**
 * Copyright (C) 2009-2010 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gdk/gdkkeysyms.h>

#include <webkit/webkit.h>

#include "wk-browser.h"
#include "wk-view.h"
#include "wk-window.h"
#include "wk-settings.h"
#include "utils.h"

enum
{
    OPEN_SIGNAL = 0,
    CLOSE_SIGNAL,
    LAST_SIGNAL
};

static guint fe_signals[LAST_SIGNAL] = { 0 };

struct _WkWindowPrivate
{
    WkWindowType type;

    GtkWidget *view;
};

static gboolean _window_on_delete_event (GtkWidget *widget,
                                         GdkEvent  *event,
                                         gpointer   user_data);

static void wk_window_class_init (WkWindowClass *klass);
static void wk_window_init (WkWindow *window);

G_DEFINE_TYPE (WkWindow, wk_window, GTK_TYPE_WINDOW);

static void
wk_window_class_init (WkWindowClass *klass)
{
    g_type_class_add_private (klass, sizeof (WkWindowPrivate));

    fe_signals[OPEN_SIGNAL] = g_signal_new ("open",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
        G_STRUCT_OFFSET (WkWindowClass, open),
        NULL,
        NULL,
        g_cclosure_marshal_VOID__VOID,
        G_TYPE_NONE,
        0);

    fe_signals[CLOSE_SIGNAL] = g_signal_new ("close",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
        G_STRUCT_OFFSET (WkWindowClass, close),
        NULL,
        NULL,
        g_cclosure_marshal_VOID__VOID,
        G_TYPE_NONE,
        0);
}

static void
wk_window_init (WkWindow *window)
{
    window->priv = G_TYPE_INSTANCE_GET_PRIVATE (window,
                                                WK_TYPE_WINDOW,
                                                WkWindowPrivate);

    window->priv->view = wk_view_new ();
    gtk_container_add (GTK_CONTAINER (window), window->priv->view);
    gtk_widget_show (GTK_WIDGET (window->priv->view));
}

static gboolean
_window_on_key_press (GtkWidget *widget,
                      GdkEventKey *event,
                      gpointer user_data)
{
    gboolean b;

    switch (event->keyval)
    {
        case GDK_w:
            if ((event->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK)
            {
                g_signal_emit_by_name (G_OBJECT (widget), "delete-event",
                                       NULL, &b);
                return TRUE;
            }
            break;

        case GDK_Escape:
            wk_browser_maybe_quit (WK_BROWSER (wk_browser_get (NULL)));
            break;

        case GDK_q:
            if ((event->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK)
            {
                wk_browser_maybe_quit (WK_BROWSER (wk_browser_get (NULL)));
            }
            break;

        case GDK_l:
            if ((event->state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK)
            {
                wk_view_url_entry_grab_focus (
                    WK_VIEW (WK_WINDOW (widget)->priv->view));
                return TRUE;
            }
            break;
    }

    return FALSE;
}

static gboolean
_window_on_delete_event (GtkWidget *widget,
                         GdkEvent  *event,
                         gpointer   user_data)
{
    WkBrowser *browser = WK_BROWSER (wk_browser_get (NULL));
    WkWindow *window = (WkWindow *) widget;
    WkWindow *main_window;

    if (window->priv->type == WK_WINDOW_TYPE_CHILD)
    {
        wk_browser_unregister_window (browser, window);
        gtk_widget_destroy (GTK_WIDGET (window));

        return TRUE;
    }

    /* if the window is the main one and an other one is open, we switch
       roles */
    if (wk_browser_has_child (browser))
    {
        wk_browser_promote_first_child (browser);
        main_window = WK_WINDOW (wk_browser_get_main_window (browser));
        main_window->priv->type = WK_WINDOW_TYPE_MAIN;

        return TRUE;
    }

    if (window->priv->type == WK_WINDOW_TYPE_MAIN)
    {
        if (!wk_browser_maybe_quit (browser))
            return TRUE;
    }

    g_object_unref (browser);

    return FALSE;
}

GtkWidget *wk_window_new (WkWindowType type,
                          gchar       *url)
{
    WkWindow *window;

    window = g_object_new (WK_TYPE_WINDOW, NULL);
    window->priv->type = type;

    if (url)
        wk_view_set_url (WK_VIEW (window->priv->view), url);

    gtk_window_set_title (GTK_WINDOW (window), "Linutop Kiosk");
    gtk_window_set_default_size (GTK_WINDOW (window), 800, 600);

    g_signal_connect (G_OBJECT (window), "delete-event",
                      G_CALLBACK (_window_on_delete_event), NULL);
    g_signal_connect (G_OBJECT (window), "key-press-event",
                      G_CALLBACK (_window_on_key_press), NULL);

    return (GTK_WIDGET (window));
}

GtkWidget *
wk_window_get_view (WkWindow *window)
{
    return window->priv->view;
}
